/**
    CS-11 Asn 2
    indp.cpp
    Purpose: Main NDP Inspector

    @author Max Schroetter
    @version 0.1 11.12.17
    Fix-Later Comments to be fixed
*/

#include "detection/detection_engine.h"
#include "events/event_queue.h"
#include "framework/inspector.h"
#include "framework/module.h"
#include "log/messages.h"
#include "profiler/profiler.h"
#include "protocols/icmp6.h"
#include "protocols/eth.h"
#include "time/packet_time.h"
#include <time.h>
#include <deque>
#include "nd.h"
#include "indp.h"

static const char* s_name = "indp";
static const char* s_help = "Inspector for Neighbor Discorvery Protocol";

static THREAD_LOCAL ProfileStats indpPerfStats;

static THREAD_LOCAL SimpleStats indpstats;

//-------------------------------------------------------------------------
//  Inspector Class
//-------------------------------------------------------------------------

class Indp : public Inspector
{
public:
    Indp(bool in_use_ipv6, long in_window_time, long in_window_count,
         long in_topology_size, MacList in_routers, PrefixList in_prefixes);

    void show(SnortConfig*) override;
    void eval(Packet*) override;
    void parse_icmp6(Packet*);
    void parse_nonND(Packet*);
    void parse_ra(Packet*, const icmp::Icmp6Hdr*, const eth::EtherHdr*,
                  const Mac_Id);
    void parse_na(Packet*, const icmp::Icmp6Hdr*, const eth::EtherHdr*,
                  const Mac_Id);
    void parse_ns(Packet*, const icmp::Icmp6Hdr*, const eth::EtherHdr*,
                  const Mac_Id);
    void parse_redirect(Packet*, const icmp::Icmp6Hdr*, const eth::EtherHdr*,
                        const Mac_Id);
    bool addr_in_prefixes(SfIp ip);
    void create_MacId(Mac_Id* macid, const uint8_t* macaddr);
    void create_HostKey( host_key* hostkey,const ip::snort_in6_addr* hostip);
    host_entry create_hostentry( const uint8_t* macaddr,
                                 const ip::snort_in6_addr* hostip, bool firstcheck);
    bool mac_changed(host_entry* host, const uint8_t* macaddr);

private:

    bool floodcontrol(Packet* p, time_t* window, long* count);
    std::deque<time_t> floodqueue;

    bool use_ipv6;
    time_t ip_flood_window_now, mld_flood_window_now;
    long flood_window_time, flood_window_count, topology_size, ip_flood_count,
         mld_flood_count ;
    MacList allowed_routers;
    RouterCache* routers;
    HostCache *hosts, *tmphosts;
    PrefixList prefixes;
};

/*
  Inspector Constructor

  @param Packetpointer Config
  @return none
  Fix-Later change Parameter to config struct
*/
Indp::Indp(bool in_use_ipv6, long in_window_time, long in_window_count,
           long in_topology_size, MacList in_routers,PrefixList in_prefixes)
{
    use_ipv6=in_use_ipv6;
    flood_window_time=in_window_time;
    flood_window_count=in_window_count;
    topology_size=in_topology_size;
    allowed_routers=in_routers;
    SfCidr onlink;
    onlink.set("fe80::0");
    onlink.set_bits(64);
    routers = new RouterCache(in_topology_size);
    hosts = new HostCache(in_topology_size);
    tmphosts = new HostCache(in_topology_size);
    in_prefixes.push_back(onlink);
    prefixes=in_prefixes;
}

bool Indp::mac_changed(host_entry* host, const uint8_t* macaddr)
{
//LogMessage("iterating: MACsrc: %hhx %hhx %hhx %hhx %hhx %hhx MAClist: %hhx %hhx %hhx %hhx %hhx %hhx \n", macaddr[0],macaddr[1],macaddr[2],macaddr[3],macaddr[4],macaddr[5],macaddr[0],macaddr[1],macaddr[2],macaddr[3],macaddr[4],macaddr[5] );
    if(memcmp(host->mac_addr,macaddr,6))
    {
        //macaddr changed event
        //memcpy( &host->mac_addr,macaddr,6);
        return true;
    }
    else
        return false;
}

void Indp::create_HostKey( host_key* hostkey,const ip::snort_in6_addr* hostip)
{
    memcpy(&hostkey->uip6addr,&hostip->u6_addr32[0],8);
    memcpy(&hostkey->lip6addr,&hostip->u6_addr32[2],8);
}

void Indp::create_MacId(Mac_Id * macid,const uint8_t* macaddr)
{
    memcpy(macid,macaddr,6);
}

host_entry Indp::create_hostentry(const uint8_t* macaddr,
                                  const ip::snort_in6_addr* hostip, bool firstcheck)
{
    host_entry newhost;
    memcpy(&newhost.mac_addr,macaddr,6);
    memcpy(&newhost.ip6_addr,hostip,16);
    newhost.last_ns=packet_time();
    newhost.last_na=packet_time();
    newhost.advertise_count=0;
    newhost.check[0]=firstcheck;
    newhost.check[1]=false;
    newhost.check[2]=false;
    return newhost;
}
/*
    Returns true if address part of one of the prefixes inside the prefix list

    @param SfIp
    @return bool
*/
bool Indp::addr_in_prefixes(SfIp ip)
{
    for(std::vector<SfCidr>::iterator it = prefixes.begin(); it != prefixes.end();
            ++it)
    {
        if (it->contains(&ip)==SfIpRet::SFIP_CONTAINS)
        {
            return true;
        }
    }
    return false;
}

/*
    Show Configuration used in verbose mode

    @param SnortConfig
    @return none
*/
void Indp::show(SnortConfig*)
{
    LogMessage("%s config:\n", s_name);
    LogMessage("\tuse_ipv6: %c \n", use_ipv6);
    LogMessage("\tflood_window_size: %ld \n", flood_window_time );
    LogMessage("\tflood_window_count: %ld \n", flood_window_count );
    LogMessage("\ttopology_size: %ld \n", topology_size );
    //Fix-Later add macs and prefix lists
}

/*
  Parses ICMP6 RouterAdvertisements Packets

  @param Packetpointer p
  @return none
*/
void Indp::parse_ra(Packet* p, const icmp::Icmp6Hdr* hdr,
                    const eth::EtherHdr* eh, const Mac_Id macid)
{
    bool mac_match=false;
    router_entry akt,old;
    const ICMP6RouterAdvertisement* ra=
        reinterpret_cast<const ICMP6RouterAdvertisement*>(hdr);
    if( ra->router_lifetime < 5 )
    {
        DetectionEngine::queue_event(INDP_GID, 3);
    }
    for(std::vector<mac_entry>::iterator it = allowed_routers.begin() ;
            it != allowed_routers.end(); ++it)
    {
        if( !memcmp((const u_char*) eh->ether_src, (const u_char*) ((*it).mac_addr),6))
        {
            mac_match=true;
        }
    }
    if(!mac_match)
    {
        DetectionEngine::queue_event(INDP_GID, 4);
        return;
    }
    else
    {
        // create router entry
        memcpy(akt.mac_addr,eh->ether_src,6);
        memcpy(&akt.src_addr,(p->ptrs.ip_api.get_ip6_src()),16);
        memcpy(&(akt.flags),eh->ether_src,1);
        // Test for ICMP Option
        if (p->ptrs.ip_api.pay_len() > 16)
        {
            uint8_t current_process_size= 16;
            const uint8_t* ra_ptr= (uint8_t*) ra;
            while (current_process_size < p->ptrs.ip_api.pay_len())
            {
                uint8_t Type = *(ra_ptr+current_process_size);
                if(Type==3)
                {
                    const ICMP6PrefixOpt* prefix_hdr=reinterpret_cast<const ICMP6PrefixOpt*>
                                                     (ra_ptr+current_process_size);
                    memcpy(&(akt.prefix_flags),&prefix_hdr->flags,1);
                    memcpy(&akt.prefix,&prefix_hdr->prefix,16);
                }
                current_process_size = current_process_size + (*(ra_ptr+current_process_size+1))
                                       *8;
            }
        }
        if(routers->exist(macid))
        {
            old = routers->get(macid);
            if ( memcmp(&(old.prefix_flags),&(akt.prefix_flags), 1))
            {
                DetectionEngine::queue_event(INDP_GID, 6);
            }
            if ( memcmp(&(old.prefix),&(akt.prefix),16))
            {
                DetectionEngine::queue_event(INDP_GID, 7);
            }
        }
        else
        {
            DetectionEngine::queue_event(INDP_GID, 5);
        }
        routers->put(macid,akt);
    }
}


/*
  Parses ICMP6 NeighborAdvertisements Packets

  @param Packetpointer p
  @return none
*/
void Indp::parse_na(Packet* p, const icmp::Icmp6Hdr* hdr,
                    const eth::EtherHdr* eh, const Mac_Id macid)
{
    const ICMP6NeighborAdvertisement* na=
        reinterpret_cast<const ICMP6NeighborAdvertisement*>(p->ptrs.icmph);
    SfIp adr,allnodes;
    host_key dst;
    adr.set( (void*) na->trg_addr.u6_addr32, AF_INET6);
    if(addr_in_prefixes(adr))
    {
        create_HostKey(&dst,&(na->trg_addr));
        allnodes.set("ff02::1");
        if(allnodes.equals(*(p->ptrs.ip_api.get_dst())))
        {
            if (!hosts->exist(dst))
            {
                if(!tmphosts->exist(dst))
                {
                    tmphosts->put(dst,create_hostentry(eh->ether_src, &na->trg_addr,false));
                }
                else
                {
                    host_entry hostentry = tmphosts->get(dst);
                    if (difftime(packet_time(), hostentry.last_ns ) < (float) 1)
                    {
                        DetectionEngine::queue_event(INDP_GID, 11);
                    }
                }
            }
            else
            {
                host_entry hostentry = hosts->get(dst);
                if (difftime(packet_time(), hostentry.last_ns ) < (float) 1)
                {
                    if (mac_changed(&hostentry,eh->ether_src))
                    {
                        DetectionEngine::queue_event(INDP_GID, 11);
                    }
                    else
                    {
                        DetectionEngine::queue_event(INDP_GID, 10);
                    }
                }
                hostentry.advertise_count+=1;
            }
        }
        host_entry tmphost;
        if (hosts->exist(dst))
        {
            tmphost= hosts->get(dst);
            if (tmphost.advertise_count > 5)
                DetectionEngine::queue_event(INDP_GID, 12);
            //LogMessage("Spoof\n");
            if (difftime(packet_time(), tmphost.last_na ) < (float) 1200)
            {
            //LogMessage("Poison\n");
            //LogMessage("iterating: MACsrc: %hhx %hhx %hhx %hhx %hhx %hhx MAClist: %hhx %hhx %hhx %hhx %hhx %hhx \n", eh->ether_src[0],eh->ether_src[1],eh->ether_src[2],eh->ether_src[3],eh->ether_src[4],eh->ether_src[5],tmphost.mac_addr[0],tmphost.mac_addr[1],tmphost.mac_addr[2],tmphost.mac_addr[3],tmphost.mac_addr[4],tmphost.mac_addr[5] );
                if (mac_changed(&tmphost,eh->ether_src))
                {
                    DetectionEngine::queue_event(INDP_GID, INDP_POSSIBLENSPOISON);
                }
                else
                {
                    tmphost.last_na=packet_time();
                }
            }
            else
            {
                //LogMessage("Time exeeded\n");
                hosts->remove(dst);
                tmphosts->put(dst,create_hostentry(eh->ether_src, &na->trg_addr,false));
            }
        }
        if (tmphosts->exist(dst))
        {
            tmphost= tmphosts->get(dst);
            if (tmphost.advertise_count > 5)
                DetectionEngine::queue_event(INDP_GID, 12);
            if (difftime(packet_time(), tmphost.last_na ) > (float) 1200)
            {
                hosts->remove(dst);
                tmphosts->put(dst,create_hostentry(eh->ether_src, &na->trg_addr,false));
            }
        }
        if (p->ptrs.ip_api.pay_len() > 24)
        {
            uint8_t current_process_size= 24;
            const uint8_t* na_ptr= (uint8_t*) na;
            while (current_process_size < p->ptrs.ip_api.pay_len())
            {
                uint8_t Type = *(na_ptr+current_process_size);
                if(Type==2)
                {
                    const ICMP6TrgOpt* Trg_hdr=reinterpret_cast<const ICMP6TrgOpt*>
                                                     (na_ptr+current_process_size);
                    if(memcmp(Trg_hdr->addr, eh->ether_src, 6)){
                        DetectionEngine::queue_event(INDP_GID, INDP_POSSIBLENSPOISON);
                    }
                }
                current_process_size = current_process_size + (*(na_ptr+current_process_size+1))
                                       *8;
            }
        }
    }
}

/*
  Parses ICMP6 NeighborSolicitations Packets
  If the node is already known in the tmpHost map and the time since the last
  neighborsolicitation is less than the last retrans timer dad fails. Link Local
  ns within the retranstimer should come from the same mac addr.


  @param Packetpointer p
  @return none
*/
void Indp::parse_ns(Packet* p, const icmp::Icmp6Hdr* hdr,
                    const eth::EtherHdr* eh, const Mac_Id macid)
{
    const ICMP6NeighborSolicitation* ns=
        reinterpret_cast<const ICMP6NeighborSolicitation*>(p->ptrs.icmph);
    ip::snort_in6_addr unspecified;
    memset(&unspecified,0,16);
    // Test if it is dad or address resolution
    if( !memcmp(p->ptrs.ip_api.get_ip6_src(),&unspecified,16))
    {
        DetectionEngine::queue_event(INDP_GID, INDP_NEWDAD_SID);
        // Test if this host is already known
        //
        // if not add him to the
        SfIp adr;
        host_key src;
        adr.set( (void*) ns->trg_addr.u6_addr32, AF_INET6);
        //LogMessage("addr:");
        //LogMessage(adr.ntoa());
        if(addr_in_prefixes(adr))
        {
            //LogMessage("\n is in Prefix\n");
            create_HostKey(&src,&ns->trg_addr);
            if (!hosts->exist(src))
            {
                //LogMessage("\n is in Prefix\n");
                if(!tmphosts->exist(src))
                {
                    //LogMessage("Created host %hhx %hhx %hhx %hhx %hhx %hhx ",(eh->ether_dst)[0],(eh->ether_dst)[1],(eh->ether_dst)[2],(eh->ether_dst)[3],(eh->ether_dst)[4],(eh->ether_dst)[5]);
                    tmphosts->put(src,create_hostentry(eh->ether_src, &ns->trg_addr,true));
                }
                else
                {
                    host_entry hostentry = tmphosts->get(src);
                    if (difftime(packet_time(), hostentry.last_ns ) < (float) 1)
                    {
                        if (mac_changed(&hostentry,eh->ether_src))
                        {
                            DetectionEngine::queue_event(INDP_GID, 11);
                        }
                        else
                        {
                            DetectionEngine::queue_event(INDP_GID, 10);
                        }
                    }
                    hostentry.last_ns=packet_time();
                    hostentry.advertise_count=0;
                    if(hostentry.check[0]==true)
                    {
                        hostentry.check[1]=true;
                    }
                    else
                    {
                        hostentry.check[0]=true;
                    }
                }
            }
            else
            {
                host_entry hostentry = hosts->get(src);
                hostentry.last_ns=packet_time();
                hostentry.advertise_count=0;
            }
        }
    }
}
/*
  Parses ICMP6 Redirect Packets

  @param Packetpointer p
  @return none
*/
void Indp::parse_redirect(Packet* p, const icmp::Icmp6Hdr* hdr,
                          const eth::EtherHdr* eh, const Mac_Id macid)
{
    const ICMP6RedirectMessage* red=
        reinterpret_cast<const ICMP6RedirectMessage*>(p->ptrs.icmph);
    if(memcmp(&red->dst_addr,&red->trg_addr,16))
    {
        bool found=false;
        for(list<pair <Mac_Id, router_entry>>::iterator it = routers->item_list.begin();
                it!= routers->item_list.end(); ++it)
        {
            if(!memcmp(&it->second.src_addr,&red->trg_addr,16))
                found=true;
        }
        if(!found)
            DetectionEngine::queue_event(INDP_GID, INDP_SPOOFEDREDIRECT_SID);
    }
}
/*
  Parses ICMP6 Packets

  @param Packetpointer p
  @return none
*/
void Indp::parse_icmp6(Packet* p)
{
    // to be sure it is the right type
    const icmp::Icmp6Hdr* hdr=reinterpret_cast<const icmp::Icmp6Hdr*>
                              (p->ptrs.icmph);
    const eth::EtherHdr* eh = layer::get_eth_layer(p);
    Mac_Id macid;
    create_MacId(&macid,eh->ether_src);
    // Only Neighbor Discovery Protocol ICMPs
    //
    // test right here if sneder or receiver is known in tmp_hosts or hosts
    if(!(hdr->type > 129 && hdr->type < 154))
    {
        return;
    }
    if(hdr->type > 132 && hdr->type < 138)
    {
        if(!floodcontrol(p,&ip_flood_window_now, &ip_flood_count))
        {
            DetectionEngine::queue_event(INDP_GID, 2);
        }
    }
    if((hdr->type > 129 && hdr->type < 133) || hdr->type == 143 )
    {
        if(!floodcontrol(p,&mld_flood_window_now, &mld_flood_count))
        {
            DetectionEngine::queue_event(INDP_GID, INDP_MLDFLOOD);
        }
    }
    if ( hdr->type == 134 )
    {
        parse_ra(p,hdr,eh,macid);
    }
    if ( hdr->type == 135 )
    {
        parse_ns(p,hdr,eh,macid);
    }
    if ( hdr->type == 136 )
    {
        parse_na(p,hdr,eh,macid);
    }
    if ( hdr->type == 137 )
    {
        parse_redirect(p,hdr,eh,macid);
    }
    if ( hdr->type == 130 || hdr->type == 151)
    {
        if (hdr-> type == 130)
        //LogMessage(" 130 or 150 \n");
        bool mac_match;
        for(std::vector<mac_entry>::iterator it = allowed_routers.begin() ;
                it != allowed_routers.end(); ++it)
        {
            //LogMessage("iterating: MACsrc: %hhx %hhx %hhx %hhx %hhx %hhx MAClist: %hhx %hhx %hhx %hhx %hhx %hhx \n", eh->ether_src[0],eh->ether_src[1],eh->ether_src[2],eh->ether_src[3],eh->ether_src[4],eh->ether_src[5],it->mac_addr[0],it->mac_addr[1],it->mac_addr[2],it->mac_addr[3],it->mac_addr[4],it->mac_addr[5] );
            if( !memcmp( eh->ether_src, it->mac_addr,6))
            {
                mac_match=true;
            }
        }
        if(!mac_match)
        {
            if (hdr->type==130)
                DetectionEngine::queue_event(INDP_GID, INDP_MLDQryNR);
            else
                DetectionEngine::queue_event(INDP_GID, INDP_MLDNONRA);
            return;
        }
    }
    if(hdr->type > 132 && hdr->type < 138)
        parse_nonND(p);
}
/*
  Checks for Packetflood

  @param Packetpointer p
  @return flase flood - true no flood
*/
bool Indp::floodcontrol(Packet* p,time_t* window, long* count)
{
    if ( difftime(packet_time(), *window ) > (float) flood_window_time )
    {
        *count = *count / 2;
        *window = packet_time();
        if ( *count > flood_window_count )
            *count = flood_window_count;
    }
    (*count)++;
    if ( *count*1.5 > flood_window_count )
        return false;
    return true;
}
/*
  Parses IP6 Packets to build faster topology

  @param Packetpointer p
  @return none
*/
void Indp::parse_nonND(Packet* p)
{
    host_key src, dst;
    const eth::EtherHdr* eh = layer::get_eth_layer(p);
    if(addr_in_prefixes(*p->ptrs.ip_api.get_src()))
    {
        create_HostKey(&src,p->ptrs.ip_api.get_ip6_src());
        if (!hosts->exist(src))
        {
            if(!tmphosts->exist(src))
            {
                //create new Entry
                tmphosts->put(src,create_hostentry(eh->ether_src, p->ptrs.ip_api.get_ip6_src(),
                                                   true));
            }
            else
            {
                host_entry* host = tmphosts->getp(src);
                //LogMessage("From Before Mac: %hhx %hhx %hhx %hhx %hhx %hhx IPV6: %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx Status: %s %s \n",(host->mac_addr)[0],(host->mac_addr)[1],(host->mac_addr)[2],(host->mac_addr)[3],(host->mac_addr)[4],(host->mac_addr)[5],(host->ip6_addr.u6_addr8)[0],host->ip6_addr.u6_addr8[1],host->ip6_addr.u6_addr8[2],host->ip6_addr.u6_addr8[3],(host->ip6_addr.u6_addr8)[4],host->ip6_addr.u6_addr8[5],host->ip6_addr.u6_addr8[6],host->ip6_addr.u6_addr8[7],(host->ip6_addr.u6_addr8)[8],host->ip6_addr.u6_addr8[9],host->ip6_addr.u6_addr8[10],host->ip6_addr.u6_addr8[11],(host->ip6_addr.u6_addr8)[12],host->ip6_addr.u6_addr8[13],host->ip6_addr.u6_addr8[14],host->ip6_addr.u6_addr8[15], (host->check[0])?"true":"false", (host->check[1])?"true":"false");
                if(host->check[0]==true)
                {
                    host->check[1]=true;
                }
                else
                {
                    host->check[0]=true;
                }
                //LogMessage("From After Mac: %hhx %hhx %hhx %hhx %hhx %hhx IPV6: %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx Status: %s %s \n",(host->mac_addr)[0],(host->mac_addr)[1],(host->mac_addr)[2],(host->mac_addr)[3],(host->mac_addr)[4],(host->mac_addr)[5],(host->ip6_addr.u6_addr8)[0],host->ip6_addr.u6_addr8[1],host->ip6_addr.u6_addr8[2],host->ip6_addr.u6_addr8[3],(host->ip6_addr.u6_addr8)[4],host->ip6_addr.u6_addr8[5],host->ip6_addr.u6_addr8[6],host->ip6_addr.u6_addr8[7],(host->ip6_addr.u6_addr8)[8],host->ip6_addr.u6_addr8[9],host->ip6_addr.u6_addr8[10],host->ip6_addr.u6_addr8[11],(host->ip6_addr.u6_addr8)[12],host->ip6_addr.u6_addr8[13],host->ip6_addr.u6_addr8[14],host->ip6_addr.u6_addr8[15], (host->check[0])?"true":"false", (host->check[1])?"true":"false");
            }
        }
    }
    if(addr_in_prefixes(*p->ptrs.ip_api.get_dst()))
    {
        create_HostKey(&dst,p->ptrs.ip_api.get_ip6_dst());
        if (!hosts->exist(dst))
        {
            if(!tmphosts->exist(dst))
            {
                tmphosts->put(src,create_hostentry(eh->ether_dst, p->ptrs.ip_api.get_ip6_dst(),
                                                   true));
            }
            else
            {
                host_entry* host = tmphosts->getp(dst);
                //LogMessage("TO After Mac: %hhx %hhx %hhx %hhx %hhx %hhx IPV6: %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx Status: %s %s \n",(host->mac_addr)[0],(host->mac_addr)[1],(host->mac_addr)[2],(host->mac_addr)[3],(host->mac_addr)[4],(host->mac_addr)[5],(host->ip6_addr.u6_addr8)[0],host->ip6_addr.u6_addr8[1],host->ip6_addr.u6_addr8[2],host->ip6_addr.u6_addr8[3],(host->ip6_addr.u6_addr8)[4],host->ip6_addr.u6_addr8[5],host->ip6_addr.u6_addr8[6],host->ip6_addr.u6_addr8[7],(host->ip6_addr.u6_addr8)[8],host->ip6_addr.u6_addr8[9],host->ip6_addr.u6_addr8[10],host->ip6_addr.u6_addr8[11],(host->ip6_addr.u6_addr8)[12],host->ip6_addr.u6_addr8[13],host->ip6_addr.u6_addr8[14],host->ip6_addr.u6_addr8[15], (host->check[0])?"true":"false", (host->check[1])?"true":"false");
                if(host->check[0]==true && host->check[1]==true )
                {
                    host->check[2]=true;
                    hosts->put(dst,*host);
                    tmphosts->remove(dst);
                }
            //if(hosts->exist(dst)) {LogMessage("\n Success \n");}
            }
        }
    }
    return;
}
/*
  API evaluation Method calls Parser

  @param Packetpointer p
  @return none
*/
void Indp::eval(Packet* p)
{
                //LogMessage("Packet IPV6 Dest: %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx ",
                 //       (p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[0],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[1],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[2],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[3],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[4],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[5],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[6],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[7],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[8],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[9],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[10],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[11],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[12],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[13],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[14],(p->ptrs.ip_api.get_ip6_dst()->u6_addr8)[15]);
                //LogMessage("Src: %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx \n",
    //                    (p->ptrs.ip_api.get_ip6_src()->u6_addr8)[0],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[1],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[2],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[3],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[4],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[5],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[6],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[7],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[8],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[9],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[10],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[11],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[12],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[13],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[14],(p->ptrs.ip_api.get_ip6_src()->u6_addr8)[15]);
    // cant register for ipv6 only
    if ( ! p->is_ip6())
        return;
    else
    {
        if( p->is_icmp() )
        {
            parse_icmp6(p);
            ++indpstats.total_packets;
        }
        else
        {
            parse_nonND(p);
            ++indpstats.total_packets;
        }
    }
}
//-------------------------------------------------------------------------
// module stuff
//-------------------------------------------------------------------------
/*
  API evaluation Method calls Parser

  @param Packetpointer p
  @return none
*/
static const Parameter router_spoof_hosts_params[] =
{
    {
        "mac", Parameter::PT_MAC, nullptr, nullptr,
        "host mac address"
    },
    {
        "mac2", Parameter::PT_MAC, nullptr, nullptr,
        "host mac address"
    },
    {
        "mac3", Parameter::PT_MAC, nullptr, nullptr,
        "host mac address"
    },
    {
        "mac4", Parameter::PT_MAC, nullptr, nullptr,
        "host mac address"
    },
    {
        "mac5", Parameter::PT_MAC, nullptr, nullptr,
        "host mac address"
    },
    { nullptr, Parameter::PT_MAX, nullptr, nullptr, nullptr }
};
/*
  API evaluation Method calls Parser

  @param Packetpointer p
  @return none
*/
static const Parameter prefix_params[] =
{
    {
        "prefix", Parameter::PT_ADDR, nullptr, nullptr,
        "addresspart of global prefixes to be tracked"
    },
    {
        "bitmask", Parameter::PT_INT, "0:128", "64",
        "Netmaks in Cidr as INT"
    },
    {
        "prefix2", Parameter::PT_ADDR, nullptr, nullptr,
        "global prefixes that should be tracked"
    },
    {
        "bitmask2", Parameter::PT_INT, "0:128", "64",
        "Netmaks in Cidr as INT"
    },
    {
        "prefix3", Parameter::PT_ADDR, nullptr, nullptr,
        "global prefixes that should be tracked"
    },
    {
        "bitmask3", Parameter::PT_INT, "0:128", "64",
        "Netmaks in Cidr as INT"
    },
    {
        "prefix4", Parameter::PT_ADDR, nullptr, nullptr,
        "global prefixes that should be tracked"
    },
    {
        "bitmask4", Parameter::PT_INT, "0:128", "64",
        "Netmaks in Cidr as INT"
    },
    {
        "prefix5", Parameter::PT_ADDR, nullptr, nullptr,
        "global prefixes that should be tracked"
    },
    {
        "bitmask5", Parameter::PT_INT, "0:128", "64",
        "Netmaks in Cidr as INT"
    },
    { nullptr, Parameter::PT_MAX, nullptr, nullptr, nullptr }
};
/*
  API evaluation Method calls Parser

  @param Packetpointer p
  @return none
*/
static const Parameter indp_params[] =
{
    { "use_ipv6", Parameter::PT_BOOL, "false", nullptr, "use_ipv6" },
    { "flood_window", Parameter::PT_INT, "1:", "5", "the flood window in second" },
    { "flood_count", Parameter::PT_INT, "1:", "10", "the count of icmp nd messages before flood is alerted" },
    { "topology_size", Parameter::PT_INT, "1:16384", "1024", "the maximum size of the topology" },
    { "routers", Parameter::PT_LIST, router_spoof_hosts_params, nullptr, "list of accepted routers in the network" },
    { "prefixes", Parameter::PT_LIST, prefix_params, nullptr, "list of global prefix that should be tracked" },
    { nullptr, Parameter::PT_MAX, nullptr, nullptr, nullptr }
};
/*
  API evaluation Method calls Parser

  @param Packetpointer p
  @return none
*/
static const RuleMap indp_rules[] =
{
    { INDP_SPOOFEDREDIRECT_SID, "Spoofed Redirect Message" },
    { INDP_FLOOD_SID, "Neighbor Discovery flood detected"},
    { INDP_ROUTERKILL_SID, "Router Kill detected"},
    { INDP_NONRA_SID, "Router Advertisement from non Router"},
    { INDP_NEWRA_SID, "New Router Alert"},
    { INDP_CHRAFLAG_SID, "Change Router Flag Alert"},
    { INDP_CHRAPREFIX_SID, "Change Router Prefix Alert"},
    { INDP_NEWDAD_SID, "New DAD"},
    { INDP_NEWHOST_SID, "New Host Alert"},
    { INDP_DADCONFLICT_SID, "DAD Conflict Alert"},
    { INDP_DADSPOOFED_SID, "Spoofed DAD Confilic Alert"},
    { INDP_UNSOLICAD_SID, "unsolicated Advertise Alert"},
    { INDP_MLDFLOOD, "MLD Flood detected"},
    { INDP_MLDNONROUTER, "MLD Query from non Router"},
    { INDP_POSSIBLENSPOISON, "Possible Neighbour Cache Poisining"},
    { INDP_MLDQryNR, "MLD Query from NON-Router MAC"},
    { INDP_MLDNONRA, "MLD Router Alert from NON-Router "},
    { 0, nullptr }
};
class IndpModule : public Module
{
public:
    IndpModule() : Module(s_name, s_help, indp_params)
    { }

    unsigned get_gid() const override
    {
        return INDP_GID;
    }

    const RuleMap* get_rules() const override
    {
        return indp_rules;
    }

    const PegInfo* get_pegs() const override
    {
        return simple_pegs;
    }

    PegCount* get_counts() const override
    {
        return (PegCount*)&indpstats;
    }

    ProfileStats* get_profile() const override
    {
        return &indpPerfStats;
    }

    bool set(const char*, Value& v, SnortConfig*) override;
    bool end(const char*, int,SnortConfig*) override;

    Usage get_usage() const override
    {
        return INSPECT;
    }

    bool use_ipv6;
    long topology_size,flood_window, flood_count;
    mac_entry entry;
    SfCidr prefix[5];
    SfCidr prefixEntry;
    int bits[5];
    MacList routers;
    PrefixList prefixes;
    bool prefixsets[5][2] = {{false,false},{false,false},{false,false},{false,false},{false,false}};
    bool macset=false;
    bool prefixset=false;

};
bool IndpModule::end(const char*, int, SnortConfig*)
{
    for (int i=0; i<5; i++)
    {
        if(prefixsets[i][0]==true && prefixsets[i][1]==true)
        {
            prefix[i].set_bits(bits[i]);
            prefixes.push_back(prefix[i]);
            prefixsets[i][0]=false;
            prefixsets[i][1]=false;
        }
        else if(prefixsets[i][0]==true || prefixsets[i][1]==true)
            return false;
    }
    return true;
}
/*
  API evaluation Method calls Parser

  @param Packetpointer p
  @return none
*/
bool IndpModule::set(const char*, Value& v, SnortConfig*)
{
    if ( v.is("use_ipv6") )
        use_ipv6 = v.get_bool();
    else if ( v.is("flood_window") )
        flood_window = v.get_long();
    else if ( v.is("flood_count") )
        flood_count = v.get_long();
    else if ( v.is("topology_size") )
        topology_size = v.get_long();
    else if ( v.is("mac") )
    {
        v.get_mac(entry.mac_addr);
        routers.push_back(entry);
    }
    else if ( v.is("mac2") )
    {
        v.get_mac(entry.mac_addr);
        routers.push_back(entry);
    }
    else if ( v.is("mac3") )
    {
        v.get_mac(entry.mac_addr);
        routers.push_back(entry);
    }
    else if ( v.is("mac4") )
    {
        v.get_mac(entry.mac_addr);
        routers.push_back(entry);
    }
    else if ( v.is("mac5") )
    {
        v.get_mac(entry.mac_addr);
        routers.push_back(entry);
    }
    else if ( v.is("prefix") )
    {
        v.get_addr(prefix[0]);
        prefixsets[0][0]=true;
    }
    else if ( v.is("prefix2") )
    {
        v.get_addr(prefix[1]);
        prefixsets[1][0]=true;
    }
    else if ( v.is("prefix3") )
    {
        v.get_addr(prefix[2]);
        prefixsets[2][0]=true;
    }
    else if ( v.is("prefix4") )
    {
        v.get_addr(prefix[3]);
        prefixsets[3][0]=true;
    }
    else if ( v.is("prefix5") )
    {
        v.get_addr(prefix[4]);
        prefixsets[4][0]=true;
    }
    else if ( v.is("bitmask") )
    {
        bits[0]=v.get_long();
        prefixsets[0][1]=true;
    }
    else if ( v.is("bitmask2") )
    {
        bits[1]=v.get_long();
        prefixsets[1][1]=true;
    }
    else if ( v.is("bitmask3") )
    {
        bits[2]=v.get_long();
        prefixsets[2][1]=true;
    }
    else if ( v.is("bitmask4") )
    {
        bits[3]=v.get_long();
        prefixsets[3][1]=true;
    }
    else if ( v.is("bitmask5") )
    {
        bits[4]=v.get_long();
        prefixsets[4][1]=true;
    }
    else
        return false;
    return true;
}
//-------------------------------------------------------------------------
// api stuff
//-------------------------------------------------------------------------
static Module* mod_ctor()
{
    return new IndpModule;
}
static void mod_dtor(Module* m)
{
    delete m;
}
static Inspector* indp_ctor(Module* m)
{
    IndpModule* mod = (IndpModule*)m;
    return new Indp(mod->use_ipv6,mod->flood_window,mod->flood_count,
                    mod->topology_size,mod->routers,mod->prefixes);
}
static void indp_dtor(Inspector* p)
{
    delete p;
}
static const InspectApi indp_api
{
    {
        PT_INSPECTOR,
        sizeof(InspectApi),
        INSAPI_VERSION,
        0,
        API_RESERVED,
        API_OPTIONS,
        s_name,
        s_help,
        mod_ctor,
        mod_dtor
    },
    IT_NETWORK,
    (uint16_t)PktType::ANY_IP,
    nullptr, // buffers
    nullptr, // service
    nullptr, // pinit
    nullptr, // pterm
    nullptr, // tinit
    nullptr, // tterm
    indp_ctor,
    indp_dtor,
    nullptr, // ssn
    nullptr  // reset
};
SO_PUBLIC const BaseApi* snort_plugins[] =
{
    &indp_api.base,
    nullptr
};
