#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "framework/ips_option.h"
#include "framework/module.h"
#include "framework/range.h"
#include "hash/sfhashfcn.h"
#include "profiler/profiler.h"
#include "protocols/packet.h"
#include "log/messages.h"

#define s_name "ip6_exthdr"

static THREAD_LOCAL ProfileStats ipExthdrPerfStats;

enum class ExtOps : uint8_t 
{ 
	None=1,
       	EQ=1, 
	NQ=0 
};

struct ExtCheck 
{
	ProtocolId pId;
	ExtOps op;
};

class IpExthdrOption : public IpsOption
{
public:
    IpExthdrOption(const ExtCheck& c) :
        IpsOption(s_name)
    { config = c; }

    uint32_t hash() const override;
    bool operator==(const IpsOption&) const override;

    IpsOption::EvalStatus eval(Cursor&, Packet*) override;

public:
    ExtCheck config;
};

//-------------------------------------------------------------------------
// class methods
//-------------------------------------------------------------------------

uint32_t IpExthdrOption::hash() const
{
    uint32_t a,b,c;

    a = (uint8_t) config.op;
    b = (uint16_t)config.pId;
    c = 0;

    mix_str(a,b,c,get_name());
    finalize(a,b,c);

    return c;
}

bool IpExthdrOption::operator==(const IpsOption& ips) const
{
    if ( strcmp(get_name(), ips.get_name()) )
        return false;

    IpExthdrOption& rhs = (IpExthdrOption&)ips;
    return ( config.pId == rhs.config.pId && config.op == rhs.config.op );
}

IpsOption::EvalStatus IpExthdrOption::eval(Cursor&, Packet* p)
{
    Profile profile(ipExthdrPerfStats);
	if(p->is_ip6()){
		int x=0;
		while( x < p->num_layers){
			if ((uint8_t)config.op){
		    	if (p->layers[x].prot_id == config.pId)
		    	        return MATCH;
		    }
		    else
		    {
		    	if (p->layers[x].prot_id == config.pId)
		    		return NO_MATCH;
		    }
		    x++;
		}
		if ((uint8_t)config.op)
		{
		    return NO_MATCH;
		}
		else
		{
		    return MATCH;
		} 
	}
	return NO_MATCH;
}

//-------------------------------------------------------------------------
// module
//-------------------------------------------------------------------------

#define s_opts "0|43|44|50|51|59|60|135"

static bool exthdr_parse(const char *str, ExtCheck* ds_ptr)
{
	//create create editable string
	char hdata[strlen(str)+1]="";
	strcpy(hdata,str);
	char * data = hdata;
	char * endptr;

	//delete quotes if forgotten by lua
	if (*data=='"') data++[strlen(data)] = 0;

	//parse operator
	if(*data=='!'|| *data=='=')
	{
		ds_ptr->op = (*data=='!') ? ExtOps::NQ : ExtOps::EQ;
		data++;
	}
	else
		ds_ptr->op = ExtOps::None;
	
	//convert to int
	long pid = strtoul( data, &endptr, 0);

	//check conversion
	if (data==endptr)
		return false;

	//set Protocol Id
	switch(pid) {
		case 0: ds_ptr->pId=ProtocolId::HOPOPTS; break;
		case 43: ds_ptr->pId= ProtocolId::ROUTING; break;
		case 44: ds_ptr->pId= ProtocolId::FRAGMENT; break;
		case 50: ds_ptr->pId= ProtocolId::ESP; break;
		case 51: ds_ptr->pId= ProtocolId::AUTH; break;
		case 59: ds_ptr->pId= ProtocolId::NONEXT; break;
		case 60: ds_ptr->pId= ProtocolId::DSTOPTS; break;
		case 135: ds_ptr->pId= ProtocolId::MOBILITY; break;
		default: return false;
	}
	return true;
}

static const Parameter s_params[] =
{
    { "~extStr", Parameter::PT_STRING, nullptr, nullptr,
      "check if the extension header by id is present or not 'id | !id | =id '" },

    { nullptr, Parameter::PT_MAX, nullptr, nullptr, nullptr }
};

#define s_help \
    "rule option to check whether the given IPv6 Extension Header is present"

class ExthdrModule : public Module
{
public:
    ExthdrModule() : Module(s_name, s_help, s_params) { }

    bool begin(const char*, int, SnortConfig*) override;
    bool set(const char*, Value&, SnortConfig*) override;

    ProfileStats* get_profile() const override
    { return &ipExthdrPerfStats; }

    ExtCheck data;
};

bool ExthdrModule::begin(const char*, int, SnortConfig*)
{
    memset(&data,0, sizeof(data));
    return true;
}

bool ExthdrModule::set(const char*, Value& v, SnortConfig*)
{
    if ( !v.is("~extStr") )
        return false;

    return exthdr_parse(v.get_string(), &data);
}

//-------------------------------------------------------------------------
// api methods
//-------------------------------------------------------------------------

// #define s_opts "0|43|44|50|51|59|60|135"

static Module* mod_ctor()
{
    return new ExthdrModule;
}

static void mod_dtor(Module* m)
{
    delete m;
}

static IpsOption* exthdr_ctor(Module* p, OptTreeNode*)
{
    ExthdrModule* m = (ExthdrModule*)p;
    return new IpExthdrOption(m->data);
}

static void exthdr_dtor(IpsOption* p)
{
    delete p;
}

static const IpsApi exthdr_api =
{
    {
        PT_IPS_OPTION,
        sizeof(IpsApi),
        IPSAPI_VERSION,
        0,
        API_RESERVED,
        API_OPTIONS,
        s_name,
        s_help,
        mod_ctor,
        mod_dtor
    },
    OPT_TYPE_DETECTION,
    1, 0,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    exthdr_ctor,
    exthdr_dtor,
    nullptr
};

SO_PUBLIC const BaseApi* snort_plugins[] =
{
	    &exthdr_api.base,
	    nullptr
};

