#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "framework/ips_option.h"
#include "framework/module.h"
#include "framework/range.h"
#include "hash/sfhashfcn.h"
#include "profiler/profiler.h"
#include "log/messages.h" 
#include "protocols/packet.h"
#include "protocols/ipv6.h"

#define s_name "ip6_optionId"

static THREAD_LOCAL ProfileStats ipOptionPerfStats;

class Ip6OptionOption : public IpsOption
{
public:
    Ip6OptionOption(const RangeCheck& c) :
        IpsOption(s_name)
    {
        config = c;
    }

    uint32_t hash() const override;
    bool operator==(const IpsOption&) const override;

    IpsOption::EvalStatus eval(Cursor&, Packet*) override;

public:
    RangeCheck config;
};

//-------------------------------------------------------------------------
// class methods
//-------------------------------------------------------------------------

uint32_t Ip6OptionOption::hash() const
{
    uint32_t a,b,c;
    a = config.op;
    b = config.min;
    c = config.max;
    mix_str(a,b,c,get_name());
    finalize(a,b,c);
    return c;
}

bool Ip6OptionOption::operator==(const IpsOption& ips) const
{
    if ( strcmp(get_name(), ips.get_name()) )
        return false;
    Ip6OptionOption& rhs = (Ip6OptionOption&)ips;
    return ( config == rhs.config );
}

//Need to be changed to option
IpsOption::EvalStatus Ip6OptionOption::eval(Cursor&, Packet* p)
{
    Profile profile(ipOptionPerfStats);
    if(p->is_ip6())
    {
        int x=0;
        while( x < p->num_layers)
        {
            if (p->layers[x].prot_id == ProtocolId::HOPOPTS ||
                    p->layers[x].prot_id == ProtocolId::DSTOPTS )
            {
                const ip::IP6Extension* oP=
                    reinterpret_cast<const ip::IP6Extension*>(p->layers[x].start);
                uint8_t len= oP->ip6e_len*8;
                const uint8_t *ptr= &(oP->ip6e_pad[0]);
                int cur_pos= -6;
                int optlen = 0, offset=0;
                while(cur_pos < len){
                    if(config.eval((long) (*(ptr+offset)))){
                        return MATCH;
                    }
                    if(*(ptr+offset) == 0){
                        cur_pos++;
                        offset++;
                    }else{
                        optlen = *((ptr+offset)+1);
                        cur_pos+=optlen;
                        offset=offset+optlen;
                    }
                }
            }

        x++;

        }
    }
    return NO_MATCH;
}


//-------------------------------------------------------------------------
// module
//-------------------------------------------------------------------------

#define RANGE "0:1048575"

static const Parameter s_params[] =
{
    {
        "~range", Parameter::PT_STRING, nullptr, nullptr,
        "check if ip option value is 'value | min<>max | <max | >min', range is " RANGE
    },

    { nullptr, Parameter::PT_MAX, nullptr, nullptr, nullptr }
};

#define s_help \
    "rule option to check ipv6 option field"

class Ip6OptionModule : public Module
{
public:
    Ip6OptionModule() : Module(s_name, s_help, s_params) { }

    bool begin(const char*, int, SnortConfig*) override;
    bool set(const char*, Value&, SnortConfig*) override;

    ProfileStats* get_profile() const override
    {
        return &ipOptionPerfStats;
    }

    RangeCheck data;
};

bool Ip6OptionModule::begin(const char*, int, SnortConfig*)
{
    data.init();
    return true;
}

bool Ip6OptionModule::set(const char*, Value& v, SnortConfig*)
{
    if ( !v.is("~range") )
        return false;
    return data.validate(v.get_string(), RANGE);
}

//-------------------------------------------------------------------------
// api methods
//-------------------------------------------------------------------------

static Module* mod_ctor()
{
    return new Ip6OptionModule;
}

static void mod_dtor(Module* m)
{
    delete m;
}

static IpsOption* option_ctor(Module* p, OptTreeNode*)
{
    Ip6OptionModule* m = (Ip6OptionModule*)p;
    return new Ip6OptionOption(m->data);
}

static void option_dtor(IpsOption* p)
{
    delete p;
}

static const IpsApi option_api =
{
    {
        PT_IPS_OPTION,
        sizeof(IpsApi),
        IPSAPI_VERSION,
        0,
        API_RESERVED,
        API_OPTIONS,
        s_name,
        s_help,
        mod_ctor,
        mod_dtor
    },
    OPT_TYPE_DETECTION,
    1, 0,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    option_ctor,
    option_dtor,
    nullptr
};

SO_PUBLIC const BaseApi* snort_plugins[] =
{
    &option_api.base,
    nullptr
};

