#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "framework/ips_option.h"
#include "framework/module.h"
#include "framework/range.h"
#include "hash/sfhashfcn.h"
#include "profiler/profiler.h"
#include "protocols/packet.h"

#define s_name "ip6_rh"

static THREAD_LOCAL ProfileStats ipRhPerfStats;

class IpRhOption : public IpsOption
{
public:
    IpRhOption(const RangeCheck& c) :
        IpsOption(s_name)
    { config = c; }

    uint32_t hash() const override;
    bool operator==(const IpsOption&) const override;

    IpsOption::EvalStatus eval(Cursor&, Packet*) override;

public:
    RangeCheck config;
};

//-------------------------------------------------------------------------
// class methods
//-------------------------------------------------------------------------

uint32_t IpRhOption::hash() const
{
    uint32_t a,b,c;

    a = config.op;
    b = config.min;
    c = config.max;

    mix_str(a,b,c,get_name());
    finalize(a,b,c);

    return c;
}

bool IpRhOption::operator==(const IpsOption& ips) const
{
    if ( strcmp(get_name(), ips.get_name()) )
        return false;

    IpRhOption& rhs = (IpRhOption&)ips;
    return ( config == rhs.config );
}

//Need to be changed to rh
IpsOption::EvalStatus IpRhOption::eval(Cursor&, Packet* p)
{
    Profile profile(ipRhPerfStats);

        if(p->is_ip6()){
                int x=0;
                while( x < p->num_layers){
                        if (p->layers[x].prot_id == ProtocolId::ROUTING)
				//const ip::IP6Extension* header = reinterpret_cast<const ip::IP6Extension*>(p->layers[x].start);
				//Accessing third 8 bits of Routing Header
				if (config.eval(reinterpret_cast<const ip::IP6Extension*>(p->layers[x].start)->ip6e_pad[0])){
                                	return MATCH;
				}
		}
	}
        return NO_MATCH;
}

//-------------------------------------------------------------------------
// module
//-------------------------------------------------------------------------

#define RANGE "0:255"

static const Parameter s_params[] =
{
    { "~range", Parameter::PT_STRING, nullptr, nullptr,
      "check if ipv6 rh type value is 'value | min<>max | <max | >min', range is " RANGE },

    { nullptr, Parameter::PT_MAX, nullptr, nullptr, nullptr }
};

#define s_help \
    "rule option to check ipv6 routing extension header type field"

class RhModule : public Module
{
public:
    RhModule() : Module(s_name, s_help, s_params) { }

    bool begin(const char*, int, SnortConfig*) override;
    bool set(const char*, Value&, SnortConfig*) override;

    ProfileStats* get_profile() const override
    { return &ipRhPerfStats; }

    RangeCheck data;
};

bool RhModule::begin(const char*, int, SnortConfig*)
{
    data.init();
    return true;
}

bool RhModule::set(const char*, Value& v, SnortConfig*)
{
    if ( !v.is("~range") )
        return false;

    return data.validate(v.get_string(), RANGE);
}

//-------------------------------------------------------------------------
// api methods
//-------------------------------------------------------------------------

static Module* mod_ctor()
{
    return new RhModule;
}

static void mod_dtor(Module* m)
{
    delete m;
}

static IpsOption* rh_ctor(Module* p, OptTreeNode*)
{
    RhModule* m = (RhModule*)p;
    return new IpRhOption(m->data);
}

static void rh_dtor(IpsOption* p)
{
    delete p;
}

static const IpsApi rh_api =
{
    {
        PT_IPS_OPTION,
        sizeof(IpsApi),
        IPSAPI_VERSION,
        0,
        API_RESERVED,
        API_OPTIONS,
        s_name,
        s_help,
        mod_ctor,
        mod_dtor
    },
    OPT_TYPE_DETECTION,
    1, 0,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    rh_ctor,
    rh_dtor,
    nullptr
};

SO_PUBLIC const BaseApi* snort_plugins[] =
{
	    &rh_api.base,
	    nullptr
};

