#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

//#include "detection/detection_defines.h"
#include "framework/ips_option.h"
#include "framework/module.h"
#include "framework/range.h"
#include "hash/sfhashfcn.h"
#include "profiler/profiler.h"
#include "protocols/packet.h"
#include "protocols/tcp.h"
#include "log/messages.h"

#include <string>
#include "protocols/protocol_ids.h"

static const char* s_name = "ipv";
static const char* s_help = "rule for ip version detection";

static THREAD_LOCAL ProfileStats ipvPerfStats;

//-------------------------------------------------------------------------
// option
//-------------------------------------------------------------------------

class IPvOption : public IpsOption
{
public:
    IPvOption(uint16_t a) : IpsOption(s_name)
    {
        this-> version = a;
    }

    bool operator==(const IpsOption&) const override;

    IpsOption::EvalStatus eval(Cursor&, Packet*) override;

private:
    uint16_t version;
};


bool IPvOption::operator==(const IpsOption& ips) const
{
    if (strcmp(s_name, ips.get_name()))
    {
        return false;
    }
    IPvOption& tbt = (IPvOption&) ips;
    return ( version == tbt.version);
}

IpsOption::EvalStatus IPvOption::eval(Cursor&, Packet* p)
{
    Profile profile(ipvPerfStats);
    IpsOption::EvalStatus rval = NO_MATCH;
    if ((((int)version == 4) and p->is_ip4()) or (((int)version == 6) and
            p->is_ip6()))
        rval = MATCH;
    return rval;
}

//-------------------------------------------------------------------------
// module
//-------------------------------------------------------------------------

static const Parameter s_params[] =
{
    {
        "~version", Parameter::PT_INT, "4:6", "4",
        "Check IP version"
    },
    { nullptr, Parameter::PT_MAX, nullptr, nullptr, nullptr }
};

class IPvModule : public Module
{
public:
    IPvModule() : Module(s_name, s_help, s_params) { }

    bool set(const char*, Value&, SnortConfig*) override;

    ProfileStats* get_profile() const override
    {
        return &ipvPerfStats;
    }

    int version=4;
};


bool IPvModule::set(const char*, Value& v, SnortConfig*)
{
    if ( v.is("~version") )
    {
        version = v.get_long();
        if (!(version==6 || version==4))
            return false;
    }
    else
        return false;
    return true;
}

//-------------------------------------------------------------------------
// api methods
//-------------------------------------------------------------------------

static Module* mod_ctor()
{
    return new IPvModule;
}

static void mod_dtor(Module* m)
{
    delete m;
}

static IpsOption* ipv_ctor(Module* p, OptTreeNode*)
{
    IPvModule* m = (IPvModule*)p;
    return new IPvOption(m->version);
}

static void ipv_dtor(IpsOption* p)
{
    delete p;
}

static const IpsApi ipv_api =
{
    {
        PT_IPS_OPTION,
        sizeof(IpsApi),
        IPSAPI_VERSION,
        0,
        API_RESERVED,
        API_OPTIONS,
        s_name,
        s_help,
        mod_ctor,
        mod_dtor
    },
    OPT_TYPE_DETECTION,
    1, PROTO_BIT__IP,
    nullptr, // pinit
    nullptr, // pterm
    nullptr, // tinit
    nullptr, // tterm
    ipv_ctor,
    ipv_dtor,
    nullptr
};

SO_PUBLIC const BaseApi* snort_plugins[] =
{
    &ipv_api.base,
    nullptr
};

